"""
Generator API.
"""
from art.estimators.generation.generator import GeneratorMixin

from art.estimators.generation.tensorflow import TensorFlowGenerator
from art.estimators.generation.tensorflow import TensorFlowV2Generator
