"""
Module containing estimators for object tracking.
"""
from art.estimators.object_tracking.object_tracker import ObjectTrackerMixin

from art.estimators.object_tracking.pytorch_goturn import PyTorchGoturn
